/*
 * mns v.80beta
 * Multifuctional Network Scanner
 * bind & aempirei 04/99
 *
 *
 * usage:
 * ./mns < <target> | <-f <file>> > [options...]
 *
 */

#include "mns.h"

extern void vuln (char *, int, int);

int usage (char *arg);

void main (int argc, char *argv[]) {
   struct utsname hinfo;
   u_short mns_opts = 0;  /* scanning options */
   int children = 30;     /* number of processes to fork */
   int opt;               /* file descriptor for getopt() */
   int i, j;
   char *scan_target;

   extern char *optarg;   /* optarg for getopt */
   extern int opterr;     /* opterr for getopt */

   setvbuf (stdout, NULL, _IONBF, 0);

   if (argc < 2) usage(argv[0]);
   
   scan_target = argv[1];

   opterr = 0; /* we like pretty output */
  
   while ((opt = getopt (argc, argv, "F:C:RVAfciqr")) != -1) {
      switch (opt) {
         case 'F': /* limit child processes */
	    mns_opts |= HOSTLIST;
	    scan_target = optarg;
	    break;
	 case 'C': /* limit child processes */
	    children = atoi(optarg);
	    break;
         case 'R': /* resolve ips to hostnames */
	    mns_opts |= RESOLVE;
            break;
	 case 'V': /* display version information */
	    puts(VERSION);
	    exit (0);
	    break;
	 case 'A': /* enable all scan checks */
	    mns_opts |= ALLVULNS;
	    break;
	 case 'q': /* qpop vuln scan */
	    mns_opts |= VULNQPOP;
	    break;
	 case 'i': /* imap vuln scan */
	    mns_opts |= VULNIMAP;
	    break;
	 case 'f': /* ftp vuln scan */
	    mns_opts |= VULNFTP;
	    break;
	 case 'c': /* cgi vuln scan */
	    mns_opts |= VULNCGI;
	    break;
	 case 'r': /* rpc vuln scan */
	    mns_opts |= VULNRPC;
	    break;
	 case '?': /* default case for unknown options */
	    usage(argv[0]);
	    break;
      }
   }

  if ((children < 2) || (children > 200)) { /* validate child process count */
      puts("error: invalid number of children");
      usage (argv[0]);
  }

  /* very cute centering technique */
  for(i = 0; i < 78; i++) putchar('*');
  i = (82 - sizeof(VERSION) - sizeof(AUTHORS)) >> 1;
  for(j = 0; j <= i; j++) putchar(' ');
  puts(VERSION " " AUTHORS);
  for(i = 0; i < 78; i++) putchar('*');
  
  uname(&hinfo);
  printf("\n\n'%s' running %s %s on a(n) %s\n\n",
         hinfo.nodename, hinfo.sysname, hinfo.release, hinfo.machine);

  /* give me root or give me death, -KILL */
  setuid (0);

  if(getuid()) {
     printf(" * %s does not have root privilages\n", getlogin());
     exit(EXIT_FAILURE); /* get lost punk */
  }

   if(mns_opts & ALLVULNS) {
      printf("Vulnerability checks: ");
      if(mns_opts & VULNQPOP) printf("qpop ");
      if(mns_opts & VULNIMAP) printf("imap ");
      if(mns_opts & VULNCGI) printf("cgi ");
      if(mns_opts & VULNRPC) printf("rpc ");
      if(mns_opts & VULNFTP) printf("ftp ");
      putchar('\n');
   }
   
   if(mns_opts & RESOLVE) printf("Resolving IPs to hostnames\n");

  vuln(scan_target, children, mns_opts);
  putchar('\n');
}

int usage (char *program) {
  int i, j;
  
  /* very cute centering technique */
  for(i = 0; i < 78; i++) putchar('*');
  i = (82 - sizeof(VERSION " " AUTHORS)) >> 1;
  for(j = 0; j <= i; j++) putchar(' ');
  puts(VERSION " " AUTHORS);
  for(i = 0; i < 78; i++) putchar('*');
  
  printf("\n usage: %s < <target> | <-F <file>> > [options...]\n", program);
  printf("   <target>         single scan where target is an IP Addresse or hostname\n"
         "   <-f <file>>      file containing a list of IP Addresses and hostnames\n"
	 "   [-qifcr]         enable vulnerability checks which consist of\n"
	 "                    the following vulnerability scanning options\n"
	 "                       -q  QPOP\n"
	 "                       -i  IMAP\n"
	 "                       -f  FTP\n"
	 "                       -c  CGI\n"
	 "                       -r  RPC\n"
	 "   [-C <children>]  set maximum child processes (2-200)\n"
	 "   [-V]             display version information\n"
	 "   [-R]             try to resolve IPs to hostnames\n"
	 "   [-A]             enable all vulerability checks\n\n");
  exit(EXIT_FAILURE);
}
