#include "mns.h"

#define OSCFGFILE "fingerprints.conf"

#define SERVICES 24
#define ACK_RAND 200
#define PACKETS 8
#define PORT_CLOSED 1

#define TH_XXX 0x40
#define TH_YYY 0x80

/* local structures */
struct fingerprint {
   u_short set:2,
           seq:2,
           ackset:3,
           urg:1;
   u_long ack;
   u_short win;
   u_short flag;
};

struct servicetype {
   char *name;
   int port;
};

typedef struct {
   u_long seq;
   u_long ack;
} hdrinfo;

/* function prototypes */
void rawsockets(int *, int *);
char *checkosmain(int, int, u_short);
void sendtcp(int, hdrinfo *, u_short);
void synscan(int, int, int *, hdrinfo);
void finscan(int, int, int *, hdrinfo);
u_long gettcp(int, struct in_addr, struct tcphdr *);

/* global variables */
struct in_addr local_ip;
time_t timedelay;
SAin sa_dest;
u_short src_port;

static struct servicetype service[SERVICES] = {
   { "echo", 7 },
   { "systat", 11 },
   { "daytime", 13 },
   { "netstat", 15 },
   { "chargen", 19 },
   { "ftp", 21 },
   { "ssh", 22 },
   { "telnet", 23 },
   { "smtp", 25 },
   { "time", 37 },
   { "domain", 53 },
   { "finger", 79 },
   { "www", 80 },
   { "kerberos", 88 },
   { "pop3", 110 },
   { "sunrcp", 111 },
   { "auth", 113 },
   { "nntp", 119 },
   { "netbios-ns", 137 },
   { "netbios-ssn", 139 },
   { "imap", 143 },
   { "login", 513 },
   { "shell", 514 },
   { "printer", 515 }
};

struct node *stealthscan (struct in_addr target,
                          struct node *walk,
                          int **plist, struct in_addr *localip)
{
   void (*scanfunc)(int, int, int *, hdrinfo);
   hdrinfo seq_ack;
   struct tcphdr itcp;
   int i, j, k;
   int sd_send, sd_recv;
   int pschkd[20] = { 0 };
   time_t timeout;
   long ack_diff = 0;
   char buffer[80];
#ifndef __linux__
   struct hostent *ent;
   struct utsname *uts;
#else
   int sockfd;
#endif

   srand (time (NULL));

   rawsockets (&sd_send, &sd_recv);

   sa_dest.sin_family = AF_INET;
   sa_dest.sin_addr = target;
   
   seq_ack.ack = 0;
   src_port = htons((rand() % 10000) + 5000);
   seq_ack.seq = rand();
   sa_dest.sin_port = ntohs((rand() % 1000) + 1000);
   
#ifndef __linux__
   uts = malloc(sizeof(struct uts));
   uname(uts);
   if((hent = gethostbyname(uts.nodename)) != NULL) {
      memcpy(&local_ip, hentry->h_addr, hentry->h_length);
      free(uts);
   } else {
      fprintf(stderr, "\n   - cannot determine local ip");
      free(uts);
      return NULL;
   }
#else
   sockfd = socket(AF_INET, SOCK_STREAM, 0);
   if (fcntl(sockfd, F_SETFL, O_NONBLOCK) == -1)
      perror ("sockfd nonblocking");
   
   signal(SIGALRM, &sighandler);
   connect(sockfd, (SA *)&sa_dest, sizeof(SAin));
   timeout = time(NULL) + TIMEOUT;
   while(timeout > time(NULL))
      if ((local_ip.s_addr = gettcp(sd_recv, sa_dest.sin_addr, &itcp)))
         break;
   close(sockfd);
#endif

   sendtcp(sd_send, &seq_ack, TH_SYN | TH_FIN);
   
   timeout = time(NULL);
   
   while(timeout + TIMEOUT > time(NULL)) {
      if ((local_ip.s_addr = gettcp(sd_recv, sa_dest.sin_addr, &itcp))) {
         if(src_port == itcp.th_dport) {
            if (ntohl(itcp.th_ack))
               ack_diff = ntohl(itcp.th_ack) - seq_ack.seq;
            break;
         }
      }
   }
   
   timedelay = time(NULL) - timeout + 2;

   if(!local_ip.s_addr) {
      fprintf(stderr, "\n   - no response from the target machine");
      return NULL;
   } else *localip = local_ip;

   if (ack_diff == 2) {
      stralloc (walk, "\n   - non-standard response, reverting to SYN scan");
      stralloc (walk, "\n   - target likely to be a Windows based machine");
      scanfunc = synscan;
   } else scanfunc = finscan;

   usleep (100);
   
   scanfunc(sd_send, sd_recv, pschkd, seq_ack);

   for (i = j = 0; i < SERVICES; i++)
     if (!pschkd[i])
       j++;

   if(j) {
      stralloc (walk, "\n   TCP ports possibly open ->");

      *plist = malloc((j + 1) * sizeof(int));
      (*plist)[0] = j;

      for (i = k = 0; i < SERVICES; i++)
	{
	  if (!pschkd[i]) {
	      k++;
	      (*plist)[k] = service[i].port;
	      sprintf (buffer, "\n      %3i   tcp\t\t%s",
		       service[i].port, service[i].name);
	      stralloc (walk, buffer);
	    }
	}
    }
  else
    *plist = NULL;

  close (sd_send);
  close (sd_recv);

  return walk;
}

struct node *checkos (struct in_addr target,
                      struct node *walk,
                      int *plist,
                      struct in_addr localip)
{
   int sd_send, sd_recv;
   char *buffer = NULL;
   int i, pcount;
   char ham_asc[2];

   if (plist != NULL) pcount = plist[0] + 1;
   else return walk;

   rawsockets (&sd_send, &sd_recv);
   local_ip = localip;

   sa_dest.sin_family = AF_INET;
   sa_dest.sin_addr = target;

   for (i = 1; i < min(pcount, 3); i++) {
      buffer = checkosmain(sd_send, sd_recv, plist[i]);
      if (buffer != NULL) {
         ham_asc[0] = buffer[0];
	 ham_asc[1] = '\0';
	 stralloc (walk, "\n   TCP Sequence Prediction Difficulty ->");
         stralloc (walk, ham_asc);
	 switch (ham_asc[0]) {
            case '0':
               stralloc (walk, "( indeterminate )");
	       break;
            case '1':
	       stralloc (walk, "( easy )");
	       break;
	    case '2':
	       stralloc (walk, "( easy )");
               break;
	    case '3':
	       stralloc (walk, "( moderate )");
	       break;
	    case '4':
	       stralloc (walk, "( moderate )");
	       break;
	    case '5':
	       stralloc (walk, "( hard )");
	       break;
	    case '6':
	       stralloc (walk, "( infeaseable )");
	       break;
	    case '7':
	       stralloc (walk, "( impossible )");
	       break;
	    case '8':
	       stralloc (walk, "( impossible )");
	       break;
	    default:
	       stralloc (walk, "( hamming error )");
	       break;
	    }
	    stralloc (walk, "\n   Operating System ->");
	    stralloc (walk, &buffer[1]);
	    break;
         }
         if (buffer != NULL)
	   break;
      }

   if (buffer != NULL)
      free (buffer);

   close (sd_send);
   close (sd_recv);

   return walk;
}

char *checkosmain (int sd_send, int sd_recv, u_short dport) {
  FILE *fs_cfg;
  char *osname;
  struct fingerprint pak[PACKETS];
  struct tcphdr tcp;
  hdrinfo seq_ack;
  u_short start, s;
  u_long myseq[PACKETS];
  u_long seq_mask = 0, last_seq = 0;
  time_t timeout;
  char line[60];
  int n, i, ham;

  for (n = 0; n < PACKETS; n++)
    bzero (&pak[n], sizeof (struct fingerprint));

  srand (time (NULL) & 0x0000ffff);
  start = s = (u_short) getpid () + 5000 + (rand () % 5000);
  seq_ack.ack = 0;
  sa_dest.sin_port = htons(dport);

#define sendtcppacket(a, b)\
   {seq_ack.seq = myseq[a] = rand();\
    src_port = htons(s + a);\
    sendtcp(sd_send, &seq_ack, b);\
    usleep(10);}

  sendtcppacket (0, TH_SYN);
  sendtcppacket (1, TH_SYN | TH_ACK);
  sendtcppacket (2, TH_FIN);
  sendtcppacket (3, TH_FIN | TH_ACK);
  sendtcppacket (4, TH_SYN | TH_FIN);
  sendtcppacket (5, TH_PUSH);
  sendtcppacket (6, TH_SYN | TH_XXX | TH_YYY);
  sendtcppacket (7, TH_SYN);	/* this extra syn is used for hamming distances
				 * an even number of responses are needed */
  timeout = time (NULL) + TIMEOUT;
  n = i = 0;

  do {
     if (gettcp (sd_recv, sa_dest.sin_addr, &tcp)) {
        if (sa_dest.sin_port == tcp.th_sport) {
           if (ntohs (tcp.th_dport) - start < PACKETS) {
              n = ntohs(tcp.th_dport) - start;
              if ((n < 0) || (n >= PACKETS)) continue;
              if (pak[n].set) continue;
              if (!n && !ntohs(tcp.th_win)) return NULL;
              pak[n].seq = tcp.th_seq != 0;
              if (pak[n].seq) {
                 seq_mask ^= tcp.th_seq;
                 last_seq = tcp.th_seq;
                 i++;
              }
              pak[n].ack = ntohl (tcp.th_ack);
              pak[n].ackset = tcp.th_ack != 0;
              if (pak[n].ackset && (myseq[n] > pak[n].ack))
                 pak[n].ackset = 2;
              if (pak[n].ackset)
                 pak[n].ack = (myseq[n] < pak[n].ack) ?
                 (pak[n].ack - myseq[n]) :
                 (myseq[n] - pak[n].ack);
              if (pak[n].ack > ACK_RAND) {
                 pak[n].ackset = 3;
                 pak[n].ack = 0;
              }
              pak[n].win = ntohs (tcp.th_win);
              pak[n].flag = tcp.th_flags;
	      pak[n].set = 1;
              pak[n].urg = tcp.th_urp != 0;
           }
        }
     }
     usleep (10);
  }
  while ((timeout > time (NULL)) && (n != PACKETS - 1));
  
  /* calculate hamming distance */
  if (i % 2) seq_mask ^= last_seq;
  ham = hamming (seq_mask);
  if (ham > 8) ham = 16 - ham;

  /* parse fingerprints */
  if ((fs_cfg = fopen (OSCFGFILE, "r")) != NULL)
    {
      struct fingerprint parse;
      int flag1 = 0, found = 0, total = 0;
      char *token;

      osname = malloc (60);

      while (fgets (line, 60, fs_cfg))
	{
	  if (line[0] == '#')
	    continue;		/* COMMENT LINE */
	  if (line[0] == '\n')
	    {			/* BLANK LINE */
	      if (flag1 && (found == total))
		{
		  fclose (fs_cfg);
		  osname[0] = (char) ham + '0';
		  return osname;
		};
	    }
	  if (line[0] == '*')
	    {			/* OS FINGERPRINT HEADER */
	      if (strlen (line) > 3)
		{
		  strcpy (osname, &line[1]);
		  osname[strlen (osname) - 1] = '\0';
		}
	      flag1 = 1;
	      total = found = 0;
	      continue;
	    }

	  /* line number */
	  token = strtok (line, " ");
	  if (isdigit (token[0]))
	    n = atoi (token);
	  else
	    {
	      flag1 = total = found = 0;
	      continue;
	    }

/*-- seq --*/
	  if ((token = strtok (NULL, " ")) != NULL)
	    parse.seq = atoi (token);
	  parse.set = isdigit (token[0]) ? 1 : 0;

/*-- ack --*/
	  if ((token = strtok (NULL, " ")) != NULL)
	    {
	      switch (token[0])
		{
		case '+':
		  parse.ackset = 1;
		  parse.ack = atol (&token[1]);
		  break;
		case '-':
		  parse.ackset = 2;
		  parse.ack = atol (&token[1]);
		  break;
		case 'R':
		  parse.ackset = 3;
		  parse.ack = 0;
		  break;
		case 'N':
		default:
		  parse.ackset = 0;
		  parse.ack = 0;
		  break;
		}
	    }

/*-- win --*/
	  if ((token = strtok (NULL, " ")) != NULL)
	    parse.win = strtol (token, NULL, 16);

/*-- flags --*/
	  if ((token = strtok (NULL, " \n")) != NULL)
	    {
	      parse.flag = 0;
	      if (strchr (token, 'S'))
		parse.flag |= TH_SYN;
	      if (strchr (token, 'R'))
		parse.flag |= TH_RST;
	      if (strchr (token, 'A'))
		parse.flag |= TH_ACK;
	      if (strchr (token, 'F'))
		parse.flag |= TH_FIN;
	      if (strchr (token, 'P'))
		parse.flag |= TH_PUSH;
	      if (strchr (token, 'X'))
		parse.flag |= TH_XXX;
	      if (strchr (token, 'Y'))
		parse.flag |= TH_YYY;
	      if (strchr (token, 'U'))
		parse.urg = 1;
	      else
		parse.urg = 0;
	    }

	  total++;

	  /* test for no response */
	  if (!pak[n].set)
	    {
	      if (parse.set)
		total = found = flag1 = 0;
	      else
		found++;
	      continue;
	    }

	  if ((parse.ackset == pak[n].ackset) &&
	      (parse.ack == pak[n].ack) &&
	      (parse.seq == pak[n].seq) &&
	      (parse.win == pak[n].win) &&
	      (parse.flag == pak[n].flag) &&
	      (parse.urg == pak[n].urg))
	    {
	      found++;
	      continue;
	    }
	  else
	    {
	      total = found = flag1 = 0;
	      continue;
	    }
	}

      fseek (fs_cfg, 0L, SEEK_END);
      fclose (fs_cfg);
      if (osname != NULL) free(osname);
      return NULL;
    }

  printf("error: cannot read from " OSCFGFILE "\n");
  return NULL;
}

/* Sends a TCP packet */
void sendtcp (int sd_send, hdrinfo * routing, u_short flags) {
  struct tcphdr tcp;

  tcp.th_sport = src_port;
  tcp.th_dport = sa_dest.sin_port;
  tcp.th_seq = htonl (routing->seq);
  tcp.th_ack = 0;
  tcp.th_x2 = 0;
  tcp.th_off = 5;
  tcp.th_flags = flags;
  tcp.th_win = htons(0x1000 | (rand () % 0x1000));
  tcp.th_urp = 0;
  tcp.th_sum = 0;
/*-- TCP Checksum --*/
#ifdef SOLARIS_CKSUM_BUG
  tcp.th_sum = sizeof (struct tcphdr);
#else
  tcp.th_sum = tcp_cksum((char *)&tcp, sizeof(struct tcphdr),
                         local_ip, sa_dest.sin_addr);
#endif /* SOLARIS_CKSUM_BUG */

  if (sendto(sd_send, &tcp, sizeof(tcp), 0, (SA *)&sa_dest, sizeof(SAin)) < 0)
     perror ("sending message");
}

#define PACKETLEN 0xffffL
#define tcphd ((struct tcphdr *)&packet[20])
#define iphd  ((struct ip *)packet)

u_long gettcp(int sd_recv, struct in_addr targetip, struct tcphdr *dtcp) {
   char packet[PACKETLEN];

   if (recv(sd_recv, packet, PACKETLEN, 0) < 0) return 0;
   if (memcmp(&targetip, &iphd->ip_src, 4)) return 0;

   memcpy(dtcp, tcphd, sizeof(struct tcphdr));

   return iphd->ip_dst.s_addr;
}

void rawsockets (int *sd_send, int *sd_recv) {
//   int on = 1;
   int rsflags;

/*-- SEND RAW socket --*/
  if ((*sd_send = socket (AF_INET, SOCK_RAW, IPPROTO_TCP)) < 0)
    {
      perror ("opening raw send socket");
      exit (EXIT_FAILURE);
    }

/*-- READ RAW socket --*/
  if ((*sd_recv = socket (AF_INET, SOCK_RAW, IPPROTO_TCP)) < 0)
    {
      perror ("opening raw read socket");
      exit (EXIT_FAILURE);
    }
  if ((rsflags = fcntl (*sd_recv, F_GETFL)) == -1)
    {
      perror ("fcntl(sd_recv,F_GETFL)");
      exit (EXIT_FAILURE);
    }
  if (fcntl (*sd_recv, F_SETFL, rsflags | O_NONBLOCK) == -1)
    {
      perror ("fcntl(sd_recv,F_SETFL)");
      exit (EXIT_FAILURE);
    }
}

void 
synscan (int sd_send, int sd_recv, int *pschkd, hdrinfo seq_ack)
{
  struct tcphdr tcp;
  int i, j;
  time_t timeout;

  for (j = 0; j < SERVICES; j++)
    pschkd[j] = PORT_CLOSED;

   for (i = 0; i < SERVICES; i++) {
      sa_dest.sin_port = htons(service[i].port);

      seq_ack.seq = rand ();
      sendtcp (sd_send, &seq_ack, TH_SYN);

      usleep (10);

      if(gettcp(sd_recv, sa_dest.sin_addr, &tcp))
         for (j = 0; j < SERVICES; j++)
            if (service[j].port == ntohs (tcp.th_sport)) {
               if (ntohs (tcp.th_win)) {
                  pschkd[j] = 0;
		  break;
               }
            }
   }

   timeout = time (NULL) + timedelay;

   while (timeout > time (NULL)) {
      if (gettcp (sd_recv, sa_dest.sin_addr, &tcp)) {
         for (j = 0; j < SERVICES; j++)
            if (service[j].port == ntohs (tcp.th_sport)) {
               if (ntohs (tcp.th_win)) {
                  pschkd[j] = 0;
                  break;
               }
            }
      }
   }
}

void finscan(int sd_send, int sd_recv, int *pschkd, hdrinfo seq_ack) {
   struct tcphdr tcp;
   int i, j;
   time_t timeout;
   u_short tcpflags = 0;

   for (i = 0; i < SERVICES; i++) {
      sa_dest.sin_port = htons(service[i].port);

      seq_ack.seq = rand();

      switch (i % 3) { /* this will help prevent scan detection */
	case 0:
	  tcpflags = 0x0;
	  break;		/* NULL */
	case 1:
	  tcpflags = TH_FIN;
	  break;		/* FIN  */
	case 2:
	  tcpflags = TH_PUSH;
	  break;		/* PUSH */
	}
      sendtcp (sd_send, &seq_ack, tcpflags);

      usleep (10);

      if (gettcp (sd_recv, sa_dest.sin_addr, &tcp))
	{
	  for (j = 0; j < SERVICES; j++)
	    if (service[j].port == ntohs (tcp.th_sport))
	      {
		pschkd[j] = PORT_CLOSED;
		break;
	      }
	}
    }
  timeout = time (NULL) + timedelay;

  while (timeout > time (NULL))
    {
      if (gettcp (sd_recv, sa_dest.sin_addr, &tcp))
	{
	  for (j = 0; j < SERVICES; j++)
	    if (service[j].port == ntohs (tcp.th_sport))
	      {
		pschkd[j] = PORT_CLOSED;
		break;
	      }
	}
    }
}
